from __future__ import annotations
from pathlib import Path

from pybi.core import DataSource
import pandas as pd

from typing import TYPE_CHECKING, Callable, Dict, List, Optional, Union, Any
import os

from pybi.core.components import (
    ContainerComponent,
    ColBoxComponent,
    ComponentTag,
    BoxComponent,
    FlowBoxComponent,
    TextComponent,
    UploadComponent,
    GridBoxComponent,
    TabsComponent,
)
from pybi.core.components.reactiveComponent import EChart, Slicer, Table, TextValue
from pybi.core.dataSource import (
    DataSourceField,
    DataSourceTable,
    DataView,
    DataViewBase,
    PivotDataView,
)
from pybi.utils.dataSourceUtils import ds2sqlite_file_base64, ds2sqlite
from pybi.utils.data_gen import (
    json_dumps_fn,
    random_ds_name,
    random_dv_name,
    get_project_root,
)
from pybi.core.sql import Sql
import pybi.utils.sql as sqlUtils
from pybi.easyEcharts.base import BaseChart

if TYPE_CHECKING:
    pass
    # from pybi.core.components import ReactiveComponent


class AppMeta:
    def __init__(self, app: App) -> None:
        self.__app = app

    def set_dbLocalStorage(self, on: bool):
        """
        是否开启数据库本地缓存
        """
        self.__app.dbLocalStorage = on
        return self


class App(ContainerComponent):
    def __init__(self) -> None:
        super().__init__(ComponentTag.App)
        self.dataSources: List[DataSource] = []
        self.dataViews: List[DataViewBase] = []
        self.__dataSetNames = set()
        self._with_temp_host_stack: List[ContainerComponent] = []
        self._clear_data = False
        self.dbLocalStorage = False
        self.__meta = AppMeta(self)

    def __record_and_check_dataset_name(self, name: str):
        if name in self.__dataSetNames:
            raise Exception(f"dataset name '{name}' is duplicate")
        self.__dataSetNames.add(name)

    @property
    def meta(self):
        return self.__meta

    def clear_all_data(self):
        self._clear_data = True

    def _get_temp_host(self):
        if self._with_temp_host_stack:
            return self._with_temp_host_stack[len(self._with_temp_host_stack) - 1]
        return None

    def set_source(self, data: pd.DataFrame, *, name: Optional[str] = None):
        name = name or random_ds_name()
        self.__record_and_check_dataset_name(name)

        ds = DataSource(name, data)
        self.dataSources.append(ds)
        return DataSourceTable(ds.name, data.columns.tolist(), host=self)

    def set_dataView(
        self,
        sql: str,
        exclude_source: Optional[List[DataSourceTable]] = None,
        *,
        name: Optional[str] = None,
    ):
        exclude_source = exclude_source or []
        name = name or random_dv_name()
        self.__record_and_check_dataset_name(name)
        dv = DataView(name, Sql(sql))

        for es in exclude_source:
            dv.exclude_source(es.source_name)

        self.dataViews.append(dv)
        return DataSourceTable(
            dv.name, sqlUtils.extract_fields_head_select(sql), host=self
        )

    def set_pivot_dataView(
        self,
        source: str,
        row: str,
        column: str,
        cell: str,
        agg="min",
        exclude_source: Optional[List[DataSourceTable]] = None,
        excludeRowFields=False,
        *,
        name: Optional[str] = None,
    ):
        exclude_source = exclude_source or []
        name = name or random_dv_name()
        self.__record_and_check_dataset_name(name)
        pdv = PivotDataView(name, source, row, column, cell, agg, excludeRowFields)

        for es in exclude_source:
            pdv.exclude_source(es.source_name)

        self.dataViews.append(pdv)
        return DataSourceTable(pdv.name, [], host=self)

    def add_upload(
        self,
        *,
        host: Optional[ContainerComponent] = None,
    ):

        cp = UploadComponent()

        host = host or self._get_temp_host() or self
        host._add_children(cp)

        return cp

    def add_text(
        self,
        text: str,
        *,
        host: Optional[ContainerComponent] = None,
    ):

        contexts = list(TextValue.extract_sql_from_text(text))

        cp = TextValue(contexts)

        host = host or self._get_temp_host() or self
        host._add_children(cp)

        return cp

    def add_slicer(
        self,
        field: Union[DataSourceField, DataSourceTable],
        *,
        host: Optional[ContainerComponent] = None,
    ):
        if isinstance(field, DataSourceTable):
            field = field[field.columns[0]]

        assert isinstance(field, DataSourceField)
        sql = f"select distinct {field._get_sql_field_name()} from {field.source_name}"
        cp = Slicer(Sql(sql))
        cp.title = field.name
        cp.add_updateInfo(field.source_name, field._get_sql_field_name())

        host = host or self._get_temp_host() or self
        host._add_children(cp)

        return cp

    def add_table(
        self,
        dataSourceTable: DataSourceTable,
        *,
        host: Optional[ContainerComponent] = None,
    ):

        sql = f"select {dataSourceTable.get_sql_fields()} from {dataSourceTable.source_name}"
        cp = Table(Sql(sql))

        host = host or self._get_temp_host() or self
        host._add_children(cp)

        return cp

    def add_echart(
        self,
        options: Union[Dict, BaseChart],
        *,
        host: Optional[ContainerComponent] = None,
    ):
        cp = None

        if isinstance(options, BaseChart):
            opts, infos = options.get_options_infos()
            cp = EChart(opts)

            for info in infos:
                cp.add_sql_path(info)
        else:
            cp = EChart(options)

            for info in EChart._extract_dataset_infos_from_option_(options):
                cp.add_sql_path(info)

        host = host or self._get_temp_host() or self
        host._add_children(cp)

        return cp

    def flowBox(
        self,
        *,
        host: Optional[ContainerComponent] = None,
    ):
        cp = FlowBoxComponent(self)

        host = host or self._get_temp_host() or self
        host._add_children(cp)

        return cp

    def gridBox(
        self,
        areas: Union[List[List[str]], str],
        *,
        host: Optional[ContainerComponent] = None,
    ):
        if isinstance(areas, str):
            areas = GridBoxComponent.areas_str2array(areas)

        cp = GridBoxComponent(areas, self)

        host = host or self._get_temp_host() or self
        host._add_children(cp)

        return cp

    def colBox(
        self,
        spec: List[int] | None = None,
        *,
        host: Optional[ContainerComponent] = None,
    ):
        cp = ColBoxComponent(spec, self)

        host = host or self._get_temp_host() or self
        host._add_children(cp)

        return cp

    def box(
        self,
        *,
        host: Optional[ContainerComponent] = None,
    ):
        cp = BoxComponent(self)
        host = host or self._get_temp_host() or self
        host._add_children(cp)
        return cp

    def add_tabs(
        self,
        names: List[str],
        mode="narrowing",
        *,
        host: Optional[ContainerComponent] = None,
    ):
        """
        mode: 'fullWidth' | 'narrowing'
            fullWidth:全宽度;页签栏会横向撑满屏幕宽度
            narrowing(默认值):收窄;页签栏向左靠拢
        """
        cp = TabsComponent(names, mode, appHost=self)
        host = host or self._get_temp_host() or self
        host._add_children(cp)
        return cp

    def save_zip_db(self, path: str):
        with open(path, mode="w", encoding="utf8") as f:
            f.write(ds2sqlite_file_base64(self.dataSources))

    def save_db(self, path: str):
        if Path(path).exists():
            os.remove(path)
        ds2sqlite(path, self.dataSources)

    def _to_json_dict(self):
        data = super()._to_json_dict()

        data["dbFile"] = ds2sqlite_file_base64(
            self.dataSources, clear_data=self._clear_data
        )
        return data

    def __reset_data(self):
        """support for run on ipython env"""
        self.children = []
        self.dataSources = []
        self.dataViews = []

    def to_json(self):
        return json_dumps_fn(self, indent=2, ensure_ascii=False)

    def to_raw_html(self):
        try:
            symbol = '"__{{__config_data__}}___"'

            config = json_dumps_fn(self, ensure_ascii=False)

            with open(
                get_project_root() / "template/index.html", mode="r", encoding="utf8"
            ) as html:
                res = html.read().replace(symbol, config)
                return res
        except Exception as e:
            raise e
        else:
            self.__reset_data()

    def to_html(self, file):
        try:
            file = Path(file)
            raw = self.to_raw_html()
            Path(file).write_text(raw, "utf8")

            print(f"to html:{file.absolute()}")
        except Exception as e:
            raise e
        else:
            self.__reset_data()



from __future__ import annotations
from typing import TYPE_CHECKING, Dict, Tuple, List, Sequence


if TYPE_CHECKING:
    from pybi.core.components.reactiveComponent import EChartDatasetInfo


class BaseChart:
    def merge(self, options: Dict):
        pass

    def get_options(self):
        return {"tooltip": {}, "legend": {}, "series": []}

    def get_options_infos(self) -> Tuple[Dict, Sequence[EChartDatasetInfo]]:
        raise NotImplementedError



from __future__ import annotations
from typing import TYPE_CHECKING, List
import sqlite3
import base64
import uuid
from pathlib import Path
from zipfile import ZipFile, ZIP_DEFLATED
import tempfile
import os


if TYPE_CHECKING:
    from pybi.core.dataSource import DataSource


def file2base64(file_path: Path):
    with open(file_path, mode="rb") as f:
        return base64.b64encode(f.read()).decode("utf8")


def _create_file_name():
    return uuid.uuid4().hex


def ds2sqlite(path: str, dataSources: List[DataSource], clear_data=False):
    with sqlite3.connect(path) as con:
        for ds in dataSources:
            data = ds._data
            if clear_data:
                data = data.head(0)
            data.to_sql(ds.name, con, if_exists="replace", index=False)


def ds2sqlite_file_base64(dataSources: List[DataSource], clear_data=False):

    filename = None
    need_remove_temp_files = False
    with tempfile.TemporaryFile() as tp:
        filename = tp.name

    # has a number name in some env
    if (filename is None) or not isinstance(filename, str):
        filename = _create_file_name()
        need_remove_temp_files = True

    ds2sqlite(filename, dataSources, clear_data)

    zipfile_name = f"{filename}.zip"

    with ZipFile(zipfile_name, "w", compression=ZIP_DEFLATED) as myzip:
        myzip.write(filename)

    res = file2base64(zipfile_name)

    if need_remove_temp_files:
        os.remove(filename)
        os.remove(zipfile_name)
    return res



from __future__ import annotations
from typing import TYPE_CHECKING, List, Union


from pybi.core.components import ComponentTag
from pybi.core.sql import Sql
from .base import ReactiveComponent
import re


if TYPE_CHECKING:
    pass

m_sql_from_text_pat = re.compile(r"(?:sql:\[_\s)(.+?)(?:\s_])", re.I)


class TextValue(ReactiveComponent):
    def __init__(self, contexts: List[Union[str, Sql]]) -> None:
        super().__init__(ComponentTag.TextValue)
        self.contexts = contexts

    @staticmethod
    def extract_sql_from_text(text: str):
        """
        >>> input = '总销售额:sql:[_ select sum(销售额) from data _]'
        >>> extract_sql_from_text(input)
        >>> ['总销售额:',Sql('select sum(销售额) from data')]
        """
        start_idx = 0

        for match in re.finditer(m_sql_from_text_pat, text):

            span = match.span()

            if span[0] > start_idx:
                # 前面有普通文本
                yield text[start_idx : span[0]]

            yield Sql(match.group(1))
            start_idx = span[1]

        end_idx = len(text) - 1

        if start_idx < end_idx:
            yield text[start_idx : len(text)]



import re

s_sql_table_name_pat = re.compile(r"(?<=from|join)\s+(\w+\b)", re.I | re.DOTALL)
s_sql_fields_head_select_pat = re.compile(
    r"(?<=select)\s+?(.+?)\s+from", re.I | re.DOTALL
)


def extract_table_names(sql: str):
    return s_sql_table_name_pat.findall(sql.strip())


def extract_fields_head_select(sql: str):
    """
    >>> extract_fields_head_select('select a as name1,b,c from (select x from tab1)')
    >>> ['name1','b','c']
    """
    res = s_sql_fields_head_select_pat.findall(sql.strip())

    each_field_has_as = (f.strip().split(" as ") for f in _iter_name(res[0]))

    each_field = (f[0] if len(f) == 1 else f[-1] for f in each_field_has_as)
    each_field = (f.strip() for f in each_field)

    return list(each_field)


def _iter_name(text: str):

    brackets_num = 0
    start_idx = 0

    for cur_idx, letter in enumerate(text):
        if letter == "," and brackets_num == 0:
            yield text[start_idx:cur_idx]
            start_idx = cur_idx + 1
        elif letter == "(":
            brackets_num += 1
        elif letter == ")":
            brackets_num -= 1

    if start_idx <= len(text) - 1:
        yield text[start_idx : len(text)]



import base64
from datetime import datetime
from enum import Enum
from pathlib import Path
from json import JSONEncoder, dumps
from functools import partial
from typing import Callable
import numpy as np
import pandas as pd
import inspect

from abc import abstractmethod


_global_id = 0


class Jsonable:
    @abstractmethod
    def _to_json_dict(self):
        data = {k: v for k, v in self.__dict__.items() if k[:1] != "_"}

        return data


def fn2str(fn: Callable):
    return inspect.getsource(fn)


def random_ds_name():
    return f"ds_{get_global_id()}"


def random_dv_name():
    return f"dv_{get_global_id()}"


def get_global_id():
    global _global_id
    _global_id += 1
    return str(_global_id)


_m_project_root = Path(__file__).absolute().parent.parent


def get_project_root():
    return _m_project_root


def _nan2None(obj):
    if isinstance(obj, dict):
        return {k: _nan2None(v) for k, v in obj.items()}
    elif isinstance(obj, list):
        return [_nan2None(v) for v in obj]
    elif isinstance(obj, float) and pd.isna(obj):
        return None
    elif isinstance(obj, np.ndarray):
        return obj.tolist()
    elif isinstance(obj, datetime):
        return str(obj)
    elif isinstance(obj, Enum):
        return obj.value
    elif isinstance(obj, Jsonable):
        return _nan2None(obj._to_json_dict())

    return obj


class _NanConverter(JSONEncoder):
    def default(self, obj):
        # possible other customizations here
        pass

    def encode(self, obj, *args, **kwargs):
        obj = _nan2None(obj)
        return super().encode(obj, *args, **kwargs)

    def iterencode(self, obj, *args, **kwargs):
        obj = _nan2None(obj)
        return super().iterencode(obj, *args, **kwargs)


json_dumps_fn = partial(dumps, cls=_NanConverter)


def data2html_img_src(data: bytes):
    b64 = base64.b64encode(data).decode("utf8")
    value = f"data:image/png;base64,{b64}"
    return value


class StrEnum(str, Enum):
    def __new__(cls, value, *args, **kwargs):
        return super().__new__(cls, value, *args, **kwargs)

    def __str__(self):
        return str(self.value)

    def _generate_next_value_(name, *_):
        return name



from .app import App

# import pybi.core.styles as styles
from pybi.core.styles import *
from pybi.easyEcharts import *
from pybi.core import Sql

# import pyvisflow.core.user_utils


app = App()

__all__ = [
    "set_source",
    "add_slicer",
    "add_table",
    "add_echart",
    "add_text",
    "colBox",
    "flowBox",
    "gridBox",
    "box",
    "app",
    "styles",
    "easy_echarts",
    "to_json",
    "sql",
    "set_dataView",
    "to_html",
    "clear_all_data",
    "add_upload",
    "_save_db",
    "meta",
    "save_zip_db",
    "add_tabs",
]

add_tabs = app.add_tabs
meta = app.meta
gridBox = app.gridBox
set_source = app.set_source
add_upload = app.add_upload
add_text = app.add_text
add_slicer = app.add_slicer
add_table = app.add_table
add_echart = app.add_echart
colBox = app.colBox
flowBox = app.flowBox
box = app.box
to_json = app.to_json
set_dataView = app.set_dataView
sql = Sql
clear_all_data = app.clear_all_data
_save_db = app.save_db
save_zip_db = app.save_zip_db

to_html = app.to_html



from __future__ import annotations
from .base import BaseChart
import pybi as pbi

from typing import TYPE_CHECKING, Optional
from pybi.core.components.reactiveComponent import EChartDatasetInfo

if TYPE_CHECKING:
    from pybi.core.dataSource import DataSourceTable


class ScatterChart(BaseChart):
    def __init__(
        self,
        data: DataSourceTable,
        x: str,
        y: str,
        color: Optional[str] = None,
        agg="avg",
    ):
        super().__init__()
        self.data = data
        self.x = x
        self.y = y
        self.color = color
        self.agg = agg
        self._series_configs = {}

    def get_options_infos(self):
        opts = super().get_options()

        opt_data = None
        if self.color:
            opt_data = pbi.set_dataView(
                f"select `{self.color}`,`{self.x}`,`{self.y}` from {self.data}"
            )
        else:
            opt_data = pbi.set_dataView(
                f"select 1,`{self.x}`,`{self.y}` from {self.data}"
            )

        series_config = {
            "type": "scatter",
        }

        series_config.update(self._series_configs)

        ds_info = EChartDatasetInfo(series_config, "dataset[0]", opt_data.source_name)

        opts["xAxis"] = [{}]

        opts["yAxis"] = [{}]

        opts["tooltip"] = {"trigger": "item"}

        return opts, [ds_info]



from __future__ import annotations
from typing import TYPE_CHECKING, List, Optional, Union


from pybi.utils.data_gen import Jsonable, get_global_id
from .componentTag import ComponentTag

if TYPE_CHECKING:
    from pybi.core.styles.styles import Style
    from pybi.core.sql import Sql


class Component(Jsonable):
    def __init__(self, tag: ComponentTag) -> None:
        self.id = get_global_id()
        self.tag = tag
        self._styles: List[Style] = []
        self.visible: Union[bool, Sql] = True
        self.__gridArea: str = ""

    def set_gridArea(self, area_name: str):
        self.__gridArea = area_name
        return self

    def _to_json_dict(self):
        data = super()._to_json_dict()
        styles_dict = {}
        for style in self._styles:
            styles_dict.update(style._get_styles_dict())
        data["styles"] = styles_dict

        if isinstance(self.visible, bool) and self.visible == True:
            del data["visible"]

        if self.__gridArea:
            data["gridArea"] = self.__gridArea

        return data

    def __add__(self, other: Style):
        self._styles.append(other)
        return self

    def __hash__(self) -> int:
        return hash(self.id)

    def __eq__(self, __o: object) -> bool:
        if isinstance(__o, self.__class__):
            return hash(__o.id) == hash(self.id)

        return False



from enum import Enum


class ComponentTag(Enum):
    App = "App"
    DataSource = "DataSource"
    Slicer = "Slicer"
    Table = "Table"
    EChart = "EChart"
    Box = "Box"
    Text = "Text"
    Upload = "Upload"
    TextValue = "TextValue"
    ColBox = "ColBox"
    FlowBox = "FlowBox"
    GridBox = "GridBox"
    Tabs = "Tabs"



from pybi.core.components.componentTag import ComponentTag
from pybi.core.components.staticComponent import TextComponent, UploadComponent
from pybi.core.components.reactiveComponent import (
    ReactiveComponent,
    Slicer,
    Table,
    TextValue,
)
from pybi.core.components.containerComponent import (
    ContainerComponent,
    BoxComponent,
    ColBoxComponent,
    FlowBoxComponent,
    GridBoxComponent,
    TabsComponent,
)



from __future__ import annotations
from typing import TYPE_CHECKING, List, Union


from pybi.core.components import ComponentTag
from pybi.core.sql import Sql
from .base import SingleReactiveComponent


if TYPE_CHECKING:
    pass


class Slicer(SingleReactiveComponent):
    def __init__(self, sql: Sql) -> None:
        super().__init__(ComponentTag.Slicer, sql)
        self.title = ""
        self.multiple = True

    def set_title(self, title: str):
        self.title = title
        return self

    def set_multiple(self, multiple: bool):
        self.multiple = multiple
        return self



import pybi.utils.sql as sqlUtils
from pybi.utils.data_gen import Jsonable


class Sql(Jsonable):
    def __init__(self, sql: str) -> None:
        self.sql = sql

    def get_table_names(self):
        return sqlUtils.extract_table_names(self.sql)

    def __str__(self) -> str:
        return f"sql:[_ {self.sql} _]"

    def _to_json_dict(self):
        data = super()._to_json_dict()
        return data
